@@ -1,4 +1,4 @@
-/* $Id: indep.c,v 1.32 2003/02/05 16:45:08 ukai Exp $ */
+/* $Id: indep.c,v 1.34 2003/07/26 17:16:24 ukai Exp $ */
 #include "fm.h"
 #include <stdio.h>
 #include <pwd.h>
@@ -7,7 +7,7 @@
 #include <stdlib.h>
 #include "indep.h"
 #include "Str.h"
-#include "gc.h"
+#include <gc.h>
 #include "myctype.h"
 #include "entity.h"
 
@@ -732,6 +732,12 @@ w3m_etc_dir()
     return w3m_dir("W3M_ETC_DIR", ETC_DIR);
 }
 
+char *
+w3m_conf_dir()
+{
+    return w3m_dir("W3M_CONF_DIR", CONF_DIR);
+}
+
 char *
 w3m_help_dir()
 {
