@@ -1,4 +1,4 @@
-/* $Id: menu.c,v 1.33 2003/07/26 17:17:28 ukai Exp $ */
+/* $Id: menu.c,v 1.40 2003/09/26 20:45:53 ukai Exp $ */
 /* 
  * w3m menu.c
  */
@@ -31,46 +31,11 @@ extern int do_getch();
 
 #ifdef USE_MENU
 
-#ifdef KANJI_SYMBOLS
-static char *FRAME[] = {
-#ifdef MENU_THIN_FRAME
-    "", "", "",
-    "", "  ", "",
-    "", "", "",
-#else				/* not MENU_THIN_FRAME */
-    "", "", "",
-    "", "  ", "",
-    "", "", "",
-#endif				/* not MENU_THIN_FRAME */
-    "", ""
-};
-#define FRAME_WIDTH 2
-
-#define G_start
- /**/
-#define G_end /**/
-#else				/* not KANJI_SYMBOLS */
-static char *N_FRAME[] = {
-    "+", "-", "+",
-    "|", " ", "|",
-    "+", "-", "+",
-    ":", ":"
-};
-#define FRAME_WIDTH 1
-
-static char *G_FRAME[] = {
-    "l", "q", "k",
-    "x", " ", "x",
-    "m", "q", "j",
-    ":", ":"
-};
-
-static char **FRAME = NULL;
+static char **FRAME;
+static int FRAME_WIDTH;
 static int graph_mode = FALSE;
-
 #define G_start  {if (graph_mode) graphstart();}
 #define G_end    {if (graph_mode) graphend();}
-#endif				/* not KANJI_SYMBOLS */
 
 static int mEsc(char c);
 static int mEscB(char c);
@@ -291,57 +256,38 @@ static int smDelTab(char c);
 /* --- MainMenu --- */
 
 static Menu MainMenu;
-#if LANG == JA
-static MenuItem MainMenuItem[] = {
-    /* type        label         variabel value func     popup keys data  */
-    {MENU_FUNC, "         (b)", NULL, 0, backBf, NULL, "b", NULL},
-    {MENU_POPUP, "Хåե (s)", NULL, 0, NULL, &SelectMenu, "s", NULL},
-    {MENU_POPUP, "     (t)", NULL, 0, NULL, &SelTabMenu, "tT", NULL},
-    {MENU_FUNC, "ɽ (v)", NULL, 0, vwSrc, NULL, "vV", NULL},
-    {MENU_FUNC, "Խ (e)", NULL, 0, editBf, NULL, "eE", NULL},
-    {MENU_FUNC, "¸ (S)", NULL, 0, svSrc, NULL, "S", NULL},
-    {MENU_FUNC, "ɤ߹   (r)", NULL, 0, reload, NULL, "rR", NULL},
-    {MENU_NOP, "", NULL, 0, nulcmd, NULL, "", NULL},
-    {MENU_FUNC, "󥯤ɽ (a)", NULL, 0, followA, NULL, "a", NULL},
-    {MENU_FUNC, "֤ɽ (n)", NULL, 0, tabA, NULL, "nN", NULL},
-    {MENU_FUNC, "󥯤¸ (A)", NULL, 0, svA, NULL, "A", NULL},
-    {MENU_FUNC, "ɽ   (i)", NULL, 0, followI, NULL, "i", NULL},
-    {MENU_FUNC, "¸   (I)", NULL, 0, svI, NULL, "I", NULL},
-    {MENU_FUNC, "ե졼ɽ (f)", NULL, 0, rFrame, NULL, "fF", NULL},
-    {MENU_NOP, "", NULL, 0, nulcmd, NULL, "", NULL},
-    {MENU_FUNC, "֥åޡ (B)", NULL, 0, ldBmark, NULL, "B", NULL},
-    {MENU_FUNC, "إ       (h)", NULL, 0, ldhelp, NULL, "hH", NULL},
-    {MENU_FUNC, "ץ   (o)", NULL, 0, ldOpt, NULL, "oO", NULL},
-    {MENU_NOP, "", NULL, 0, nulcmd, NULL, "", NULL},
-    {MENU_FUNC, "λ         (q)", NULL, 0, qquitfm, NULL, "qQ", NULL},
-    {MENU_END, "", NULL, 0, nulcmd, NULL, "", NULL},
-};
-#else				/* LANG != JA */
+#ifdef USE_M17N
+/* FIXME: gettextize here */
+static wc_ces MainMenuCharset = WC_CES_US_ASCII;	/* FIXME: charset of source code */
+static int MainMenuEncode = FALSE;
+#endif
+
 static MenuItem MainMenuItem[] = {
     /* type        label           variable value func     popup keys data  */
-    {MENU_FUNC, " Back         (b) ", NULL, 0, backBf, NULL, "b", NULL},
-    {MENU_POPUP, " Select Buffer(s) ", NULL, 0, NULL, &SelectMenu, "s", NULL},
-    {MENU_POPUP, " Select Tab   (t) ", NULL, 0, NULL, &SelTabMenu, "tT", NULL},
-    {MENU_FUNC, " View Source  (v) ", NULL, 0, vwSrc, NULL, "vV", NULL},
-    {MENU_FUNC, " Edit Source  (e) ", NULL, 0, editBf, NULL, "eE", NULL},
-    {MENU_FUNC, " Save Source  (S) ", NULL, 0, svSrc, NULL, "S", NULL},
-    {MENU_FUNC, " Reload       (r) ", NULL, 0, reload, NULL, "rR", NULL},
-    {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, "", NULL},
-    {MENU_FUNC, " Go Link      (a) ", NULL, 0, followA, NULL, "a", NULL},
-    {MENU_FUNC, "   on New Tab (n) ", NULL, 0, tabA, NULL, "nN", NULL},
-    {MENU_FUNC, " Save Link    (A) ", NULL, 0, svA, NULL, "A", NULL},
-    {MENU_FUNC, " View Image   (i) ", NULL, 0, followI, NULL, "i", NULL},
-    {MENU_FUNC, " Save Image   (I) ", NULL, 0, svI, NULL, "I", NULL},
-    {MENU_FUNC, " View Frame   (f) ", NULL, 0, rFrame, NULL, "fF", NULL},
-    {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, "", NULL},
-    {MENU_FUNC, " Bookmark     (B) ", NULL, 0, ldBmark, NULL, "B", NULL},
-    {MENU_FUNC, " Help         (h) ", NULL, 0, ldhelp, NULL, "hH", NULL},
-    {MENU_FUNC, " Option       (o) ", NULL, 0, ldOpt, NULL, "oO", NULL},
-    {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, "", NULL},
-    {MENU_FUNC, " Quit         (q) ", NULL, 0, qquitfm, NULL, "qQ", NULL},
+    {MENU_FUNC, N_(" Back         (b) "), NULL, 0, backBf, NULL, "b", NULL},
+    {MENU_POPUP, N_(" Select Buffer(s) "), NULL, 0, NULL, &SelectMenu, "s",
+     NULL},
+    {MENU_POPUP, N_(" Select Tab   (t) "), NULL, 0, NULL, &SelTabMenu, "tT",
+     NULL},
+    {MENU_FUNC, N_(" View Source  (v) "), NULL, 0, vwSrc, NULL, "vV", NULL},
+    {MENU_FUNC, N_(" Edit Source  (e) "), NULL, 0, editBf, NULL, "eE", NULL},
+    {MENU_FUNC, N_(" Save Source  (S) "), NULL, 0, svSrc, NULL, "S", NULL},
+    {MENU_FUNC, N_(" Reload       (r) "), NULL, 0, reload, NULL, "rR", NULL},
+    {MENU_NOP, N_(" ---------------- "), NULL, 0, nulcmd, NULL, "", NULL},
+    {MENU_FUNC, N_(" Go Link      (a) "), NULL, 0, followA, NULL, "a", NULL},
+    {MENU_FUNC, N_("   on New Tab (n) "), NULL, 0, tabA, NULL, "nN", NULL},
+    {MENU_FUNC, N_(" Save Link    (A) "), NULL, 0, svA, NULL, "A", NULL},
+    {MENU_FUNC, N_(" View Image   (i) "), NULL, 0, followI, NULL, "i", NULL},
+    {MENU_FUNC, N_(" Save Image   (I) "), NULL, 0, svI, NULL, "I", NULL},
+    {MENU_FUNC, N_(" View Frame   (f) "), NULL, 0, rFrame, NULL, "fF", NULL},
+    {MENU_NOP, N_(" ---------------- "), NULL, 0, nulcmd, NULL, "", NULL},
+    {MENU_FUNC, N_(" Bookmark     (B) "), NULL, 0, ldBmark, NULL, "B", NULL},
+    {MENU_FUNC, N_(" Help         (h) "), NULL, 0, ldhelp, NULL, "hH", NULL},
+    {MENU_FUNC, N_(" Option       (o) "), NULL, 0, ldOpt, NULL, "oO", NULL},
+    {MENU_NOP, N_(" ---------------- "), NULL, 0, nulcmd, NULL, "", NULL},
+    {MENU_FUNC, N_(" Quit         (q) "), NULL, 0, qquitfm, NULL, "qQ", NULL},
     {MENU_END, "", NULL, 0, nulcmd, NULL, "", NULL},
 };
-#endif				/* LANG != JA  */
 
 /* --- MainMenu (END) --- */
 
@@ -389,7 +335,7 @@ new_menu(Menu *menu, MenuItem *item)
 		p++;
 	    }
 	}
-	l = strlen(item[i].label);
+	l = get_strwidth(item[i].label);
 	if (l > menu->width)
 	    menu->width = l;
     }
@@ -452,43 +398,31 @@ draw_menu(Menu *menu)
     w = menu->width + 2 * FRAME_WIDTH;
     y = menu->y - 1;
 
-#ifndef KANJI_SYMBOLS
-    if (FRAME == NULL) {
-	if (graph_ok()) {
-	    graph_mode = TRUE;
-	    FRAME = G_FRAME;
-	}
-	else {
-	    FRAME = N_FRAME;
-	}
-    }
-#endif				/* not KANJI_SYMBOLS */
-
     if (menu->offset == 0) {
 	G_start;
-	mvaddstr(y, x, FRAME[0]);
+	mvaddstr(y, x, FRAME[3]);
 	for (i = FRAME_WIDTH; i < w - FRAME_WIDTH; i += FRAME_WIDTH)
-	    mvaddstr(y, x + i, FRAME[1]);
-	mvaddstr(y, x + i, FRAME[2]);
+	    mvaddstr(y, x + i, FRAME[10]);
+	mvaddstr(y, x + i, FRAME[6]);
 	G_end;
     }
     else {
 	G_start;
-	mvaddstr(y, x, FRAME[3]);
+	mvaddstr(y, x, FRAME[5]);
 	G_end;
-	for (i = FRAME_WIDTH; i < w - FRAME_WIDTH; i += FRAME_WIDTH)
-	    mvaddstr(y, x + i, FRAME[4]);
+	for (i = FRAME_WIDTH; i < w - FRAME_WIDTH; i++)
+	    mvaddstr(y, x + i, " ");
 	G_start;
 	mvaddstr(y, x + i, FRAME[5]);
 	G_end;
 	i = (w / 2 - 1) / FRAME_WIDTH * FRAME_WIDTH;
-	mvaddstr(y, x + i, FRAME[9]);
+	mvaddstr(y, x + i, ":");
     }
 
     for (j = 0; j < menu->height; j++) {
 	y++;
 	G_start;
-	mvaddstr(y, x, FRAME[3]);
+	mvaddstr(y, x, FRAME[5]);
 	G_end;
 	draw_menu_item(menu, menu->offset + j);
 	G_start;
@@ -498,23 +432,23 @@ draw_menu(Menu *menu)
     y++;
     if (menu->offset + menu->height == menu->nitem) {
 	G_start;
-	mvaddstr(y, x, FRAME[6]);
+	mvaddstr(y, x, FRAME[9]);
 	for (i = FRAME_WIDTH; i < w - FRAME_WIDTH; i += FRAME_WIDTH)
-	    mvaddstr(y, x + i, FRAME[7]);
-	mvaddstr(y, x + i, FRAME[8]);
+	    mvaddstr(y, x + i, FRAME[10]);
+	mvaddstr(y, x + i, FRAME[12]);
 	G_end;
     }
     else {
 	G_start;
-	mvaddstr(y, x, FRAME[3]);
+	mvaddstr(y, x, FRAME[5]);
 	G_end;
-	for (i = FRAME_WIDTH; i < w - FRAME_WIDTH; i += FRAME_WIDTH)
-	    mvaddstr(y, x + i, FRAME[4]);
+	for (i = FRAME_WIDTH; i < w - FRAME_WIDTH; i++)
+	    mvaddstr(y, x + i, " ");
 	G_start;
 	mvaddstr(y, x + i, FRAME[5]);
 	G_end;
 	i = (w / 2 - 1) / FRAME_WIDTH * FRAME_WIDTH;
-	mvaddstr(y, x + i, FRAME[10]);
+	mvaddstr(y, x + i, ":");
     }
 }
 
@@ -762,6 +696,28 @@ new_option_menu(Menu *menu, char **label, int *variable, void (*func) ())
     new_menu(menu, item);
 }
 
+static void
+set_menu_frame(void)
+{
+    if (graph_ok()) {
+	graph_mode = TRUE;
+	FRAME_WIDTH = 1;
+	FRAME = graph_symbol;
+    }
+    else {
+	graph_mode = FALSE;
+#ifdef USE_M17N
+	FRAME_WIDTH = 0;
+	FRAME = get_symbol(DisplayCharset, &FRAME_WIDTH);
+	if (!WcOption.use_wide)
+	    FRAME_WIDTH = 1;
+#else
+	FRAME_WIDTH = 1;
+	FRAME = get_symbol();
+#endif
+    }
+}
+
 /* --- MenuFunctions --- */
 
 #ifdef __EMX__
@@ -998,10 +954,15 @@ menu_search_forward(Menu *menu, int from)
     if (str == NULL || *str == '\0')
 	return -1;
     SearchString = str;
+#ifdef USE_M17N
+    if (SearchConv && !WcOption.pre_conv &&
+	Currentbuf->document_charset != DisplayCharset)
+	str = wtf_conv_fit(str, Currentbuf->document_charset);
+#endif
     menuSearchRoutine = menuForwardSearch;
-    found = menuForwardSearch(menu, SearchString, from + 1);
+    found = menuForwardSearch(menu, str, from + 1);
     if (WrapSearch && found == -1)
-	found = menuForwardSearch(menu, SearchString, 0);
+	found = menuForwardSearch(menu, str, 0);
     if (found >= 0)
 	return found;
     disp_message("Not found", TRUE);
@@ -1045,12 +1006,17 @@ menu_search_backward(Menu *menu, int from)
     if (str != NULL && *str == '\0')
 	str = SearchString;
     if (str == NULL || *str == '\0')
-	return (MENU_NOTHING);
+	return -1;
     SearchString = str;
+#ifdef USE_M17N
+    if (SearchConv && !WcOption.pre_conv &&
+	Currentbuf->document_charset != DisplayCharset)
+	str = wtf_conv_fit(str, Currentbuf->document_charset);
+#endif
     menuSearchRoutine = menuBackwardSearch;
-    found = menuBackwardSearch(menu, SearchString, from - 1);
+    found = menuBackwardSearch(menu, str, from - 1);
     if (WrapSearch && found == -1)
-	found = menuBackwardSearch(menu, SearchString, menu->nitem);
+	found = menuBackwardSearch(menu, str, menu->nitem);
     if (found >= 0)
 	return found;
     disp_message("Not found", TRUE);
@@ -1073,20 +1039,26 @@ menu_search_next_previous(Menu *menu, int from, int reverse)
     int found;
     static int (*routine[2]) (Menu *, char *, int) = {
     menuForwardSearch, menuBackwardSearch};
+    char *str;
 
     if (menuSearchRoutine == NULL) {
 	disp_message("No previous regular expression", TRUE);
 	return -1;
     }
+#ifdef USE_M17N
+    str = SearchString;
+    if (SearchConv && !WcOption.pre_conv &&
+	Currentbuf->document_charset != DisplayCharset)
+	str = wtf_conv_fit(str, Currentbuf->document_charset);
+#endif
     if (reverse != 0)
 	reverse = 1;
     if (menuSearchRoutine == menuBackwardSearch)
 	reverse ^= 1;
     from += reverse ? -1 : 1;
-    found = (*routine[reverse]) (menu, SearchString, from);
+    found = (*routine[reverse]) (menu, str, from);
     if (WrapSearch && found == -1)
-	found =
-	    (*routine[reverse]) (menu, SearchString, reverse * menu->nitem);
+	found = (*routine[reverse]) (menu, str, reverse * menu->nitem);
     if (found >= 0)
 	return found;
     disp_message("Not found", TRUE);
@@ -1311,6 +1283,8 @@ mMouse(char c)
 void
 popupMenu(int x, int y, Menu *menu)
 {
+    set_menu_frame();
+
     initSelectMenu();
     initSelTabMenu();
 
@@ -1328,8 +1302,7 @@ mainMenu(int x, int y)
     popupMenu(x, y, &MainMenu);
 }
 
-void
-mainMn(void)
+DEFUN(mainMn, MAIN_MENU MENU, "Popup menu")
 {
     Menu *menu = &MainMenu;
     char *data;
@@ -1357,8 +1330,7 @@ mainMn(void)
 
 /* --- SelectMenu --- */
 
-void
-selMn(void)
+DEFUN(selMn, SELECT_MENU, "Popup buffer selection menu")
 {
     int x = Currentbuf->cursorX + Currentbuf->rootX,
 	y = Currentbuf->cursorY + Currentbuf->rootY;
@@ -1417,7 +1389,7 @@ initSelectMenu(void)
 	if (len < str->length)
 	    len = str->length;
     }
-    l = strlen(comment);
+    l = get_strwidth(comment);
     if (len < l + 4)
 	len = l + 4;
     if (len > COLS - 2 * FRAME_WIDTH)
@@ -1505,8 +1477,7 @@ smDelBuf(char c)
 
 /* --- SelTabMenu --- */
 
-void
-tabMn(void)
+DEFUN(tabMn, TAB_MENU, "Popup tab selection menu")
 {
     int x = Currentbuf->cursorX + Currentbuf->rootX,
 	y = Currentbuf->cursorY + Currentbuf->rootY;
@@ -1655,6 +1626,8 @@ optionMenu(int x, int y, char **label, int *variable, int initial,
 {
     Menu menu;
 
+    set_menu_frame();
+
     new_option_menu(&menu, label, variable, func);
     menu.cursorX = COLS - 1;
     menu.cursorY = LASTLINE;
@@ -1676,6 +1649,9 @@ interpret_menu(FILE * mf)
     char *p, *s;
     int in_menu = 0, nmenu = 0, nitem = 0, type;
     MenuItem *item = NULL;
+#ifdef USE_M17N
+    wc_ces charset = SystemCharset;
+#endif
 
     while (!feof(mf)) {
 	line = Strfgets(mf);
@@ -1683,6 +1659,9 @@ interpret_menu(FILE * mf)
 	Strremovefirstspaces(line);
 	if (line->length == 0)
 	    continue;
+#ifdef USE_M17N
+	line = wc_Str_conv(line, charset, InnerCharset);
+#endif
 	p = line->ptr;
 	s = getWord(&p);
 	if (*s == '#')		/* comment */
@@ -1700,9 +1679,7 @@ interpret_menu(FILE * mf)
 		item[nitem].type = MENU_END;
 	    }
 	}
-	else {
-	    if (strcmp(s, "menu"))	/* error */
-		continue;
+	else if (!strcmp(s, "menu")) {
 	    s = getQWord(&p);
 	    if (*s == '\0')	/* error */
 		continue;
@@ -1715,6 +1692,14 @@ interpret_menu(FILE * mf)
 	    nitem = 0;
 	    item[nitem].type = MENU_END;
 	}
+#ifdef USE_M17N
+	else if (!strcmp(s, "charset") || !strcmp(s, "encoding")) {
+	    s = getQWord(&p);
+	    if (*s == '\0')	/* error */
+		continue;
+	    charset = wc_guess_charset(s, charset);
+	}
+#endif
     }
 }
 
@@ -1736,6 +1721,20 @@ initMenu(void)
     w3mMenuList[2].item = NULL;
     w3mMenuList[3].id = NULL;
 
+#ifdef USE_M17N
+    if (!MainMenuEncode) {
+	MenuItem *item;
+#if ENABLE_NLS
+	/* FIXME: charset that gettext(3) returns */
+	MainMenuCharset = SystemCharset;
+#endif
+	for (item = MainMenuItem; item->type != MENU_END; item++)
+	    item->label =
+		wc_conv(gettext(item->label), MainMenuCharset,
+			InnerCharset)->ptr;
+	MainMenuEncode = TRUE;
+    }
+#endif
     if ((mf = fopen(confFile(MENU_FILE), "rt")) != NULL) {
 	interpret_menu(mf);
 	fclose(mf);
@@ -1860,7 +1859,7 @@ link_menu(Buffer *buf)
 	if (!l->url)
 	    p = "";
 	else if (DecodeURL)
-	    p = url_unquote_conv(l->url, buf->document_code);
+	    p = url_unquote_conv(l->url, buf->document_charset);
 	else
 	    p = l->url;
 	Strcat_charp(str, p);
@@ -1870,6 +1869,7 @@ link_menu(Buffer *buf)
     }
     label[nitem] = NULL;
 
+    set_menu_frame();
     new_option_menu(&menu, label, &linkV, NULL);
 
     menu.initial = 0;
@@ -2037,6 +2037,8 @@ list_menu(Buffer *buf)
     }
     label[nitem] = NULL;
 
+    set_menu_frame();
+    set_menu_frame();
     new_option_menu(&menu, label, &key, NULL);
 
     menu.initial = 0;
