@@ -1,4 +1,4 @@
-/* $Id: url.c,v 1.82 2003/09/22 17:37:41 ukai Exp $ */
+/* $Id: url.c,v 1.87 2003/12/08 16:17:21 ukai Exp $ */
 #include "fm.h"
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -22,9 +22,9 @@
 
 #ifdef USE_SSL
 #ifndef SSLEAY_VERSION_NUMBER
-#include <crypto.h>		/* SSLEAY_VERSION_NUMBER may be here */
+#include <openssl/crypto.h>		/* SSLEAY_VERSION_NUMBER may be here */
 #endif
-#include <err.h>
+#include <openssl/err.h>
 #endif
 
 #ifdef	__WATT32__
@@ -248,7 +248,7 @@ free_ssl_ctx()
 }
 
 #if SSLEAY_VERSION_NUMBER >= 0x00905100
-#include <rand.h>
+#include <openssl/rand.h>
 static void
 init_PRNG()
 {
@@ -378,6 +378,7 @@ openSSLHandle(int sock, char *hostname, char **p_cert)
     close(sock);
     if (handle)
 	SSL_free(handle);
+    /* FIXME: gettextize? */
     disp_err_message(Sprintf
 		     ("SSL error: %s",
 		      ERR_error_string(ERR_get_error(), NULL))->ptr, FALSE);
@@ -454,6 +455,7 @@ openSocket(char *const hostname,
     MySignalHandler(*volatile prevtrap) (SIGNAL_ARG) = NULL;
 
     if (fmInitialized) {
+	/* FIXME: gettextize? */
 	message(Sprintf("Opening socket...")->ptr, 0, 0);
 	refresh();
     }
@@ -999,8 +1001,19 @@ parseURL2(char *url, ParsedURL *pu, ParsedURL *current)
 	}
 	return;
     }
-    if (pu->scheme == SCM_LOCAL)
-	pu->file = file_quote(expandName(file_unquote(pu->file)));
+    if (pu->scheme == SCM_LOCAL) {
+	char *q = expandName(file_unquote(pu->file));
+#ifdef SUPPORT_DOS_DRIVE_PREFIX
+	Str drive;
+	if (IS_ALPHA(q[0]) && q[1] == ':') {
+	    drive = Strnew_charp_n(q, 2);
+	    Strcat_charp(drive, file_quote(q+2));
+	    pu->file = drive->ptr;
+	}
+	else
+#endif
+	    pu->file = file_quote(q);
+    }
 
     if (current && (pu->scheme == current->scheme ||
 		    (pu->scheme == SCM_FTP && current->scheme == SCM_FTPDIR) ||
@@ -1236,7 +1249,7 @@ getURLScheme(char **url)
     int i;
     int scheme = SCM_MISSING;
 
-    while (*p && (IS_ALPHA(*p) || *p == '.' || *p == '+' || *p == '-'))
+    while (*p && (IS_ALNUM(*p) || *p == '.' || *p == '+' || *p == '-'))
 	p++;
     if (*p == ':') {		/* scheme found */
 	scheme = SCM_UNKNOWN;
@@ -1338,7 +1351,10 @@ HTTPrequestURI(ParsedURL *pu, HRequest *hr)
 	}
     }
     else {
+	char *save_label = pu->label;
+	pu->label = NULL;
 	Strcat(tmp, _parsedURL2Str(pu, TRUE));
+	pu->label = save_label;
     }
     return tmp;
 }
@@ -1625,7 +1641,6 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 #endif				/* USE_SSL */
 		non_null(HTTP_proxy)) && !Do_not_use_proxy &&
 	    pu->host != NULL && !check_no_proxy(pu->host)) {
-	    char *save_label;
 	    hr->flag |= HR_FLAG_PROXY;
 #ifdef USE_SSL
 	    if (pu->scheme == SCM_HTTPS && *status == HTST_CONNECT) {
@@ -1657,8 +1672,6 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 #endif
 		return uf;
 	    }
-	    save_label = pu->label;
-	    pu->label = NULL;
 #ifdef USE_SSL
 	    if (pu->scheme == SCM_HTTPS) {
 		if (*status == HTST_NORMAL) {
@@ -1677,7 +1690,6 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 	    {
 		tmp = HTTPrequest(pu, current, hr, extra_header);
 		*status = HTST_NORMAL;
-		pu->label = save_label;
 	    }
 	}
 	else {
